from .ApiService import ApiService
from . import Configuration
import os
import yaml
from enum import Enum
import time
import logging
from .Dbfs import dbfs_upload

endpoint = "clusters"
logging.basicConfig(
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)

_api_service = ApiService(Configuration.config)


class ClusterState(Enum):
    PENDING = 1
    RUNNING = 2
    RESTARTING = 3
    RESIZING = 4
    TERMINATING = 5
    TERMINATED = 6
    ERROR = 7
    UNKNOWN = 8


class ClusterAction(Enum):
    PIN = 1
    UNPIN = 2
    START = 3
    RESTART = 4
    STOP = 5
    DELETE = 6


def cluster_name_exists(name: str):

    found = cluster_get_name_ids(name)

    return len(found) > 0


def cluster_get_by_name(name: str):

    response = cluster_list()
    found = []
    if response:
        found = [c for c in response["clusters"] if c["cluster_name"] == name]

    return found


def cluster_action(cluster_id: str, cluster_action: ClusterAction):

    action: str

    if cluster_action == ClusterAction.PIN:
        action = "pin"

    elif cluster_action == ClusterAction.UNPIN:
        action = "unpin"

    elif cluster_action == ClusterAction.START:
        action = "start"

    elif cluster_action == ClusterAction.RESTART:
        action = "restart"

    elif cluster_action == ClusterAction.STOP:
        action = "delete"

    elif cluster_action == ClusterAction.DELETE:
        action = "permanent-delete"

    cluster = dict()
    cluster["cluster_id"] = cluster_id

    logger.info(f"{cluster_action.name} cluster: {cluster_id}")

    response = _api_service.api_post(endpoint, action, cluster)

    return response


def clusters_create(
    cluster_defn_folder: str,
    pin: bool = True,
    stop: bool = True,
    delete_if_exists: bool = False,
    allow_duplicate_names: bool = False,
    init_script_path: str = None,
):

    directory = os.path.abspath(cluster_defn_folder)
    for entry in os.scandir(directory):
        if entry.path.endswith(".yaml") and entry.is_file():

            response = cluster_create(
                entry.path,
                pin,
                stop,
                delete_if_exists,
                allow_duplicate_names,
                init_script_path,
            )


def cluster_create(
    cluster_defn_path: str,
    pin: bool = True,
    stop: bool = True,
    delete_if_exists: bool = False,
    allow_duplicate_names: bool = False,
    init_script_path: str = None,
):

    with open(cluster_defn_path, "r") as f:
        cluster_defn: dict = yaml.safe_load(f)

    cluster_name: str = cluster_defn["cluster_name"]

    # upload init script
    if cluster_defn.get("init_scripts"):

        for i in cluster_defn["init_scripts"]:

            if i.get("dbfs"):

                to_path = i["dbfs"]["destination"]
                from_path = os.path.abspath(f"{init_script_path}/{cluster_name}.sh")
                dbfs_upload(from_path, to_path, overwrite=True)
                logger.info(
                    f"Uploaded cluster {cluster_name} init script: {from_path} => {to_path}"
                )

    logger.info(
        f"Attempting to create a cluster using the declaration at path {cluster_defn_path} with the name {cluster_name}"
    )

    clusters = cluster_get_by_name(cluster_name)
    logger.info(f"Found {len(clusters)} clusters with the name {cluster_name}")

    if len(clusters) > 0 and delete_if_exists:

        logger.info(
            f"Cluster(s) named {cluster_name} already exist. Deleting existing clusters."
        )
        cluster_delete_clusters(clusters)

    if len(clusters) == 0 or delete_if_exists or allow_duplicate_names:

        logger.info(f"Creating cluster {cluster_name}")
        response = _api_service.api_post(endpoint, "create", cluster_defn)
        cluster_id = response["cluster_id"]
        logger.info(f"Cluster {cluster_name} created with id: {cluster_id}")
        create_response = {"cluster_id": cluster_id}

        if pin:
            cluster_action(cluster_id, ClusterAction.PIN)
        if stop:
            cluster_action(cluster_id, ClusterAction.STOP)
        create_response = {"clusters": [create_response], "created": True}

    else:

        logger.info(
            f"Cluster {cluster_name} already exists and delete_if_exists is disabled."
        )
        create_response = {"clusters": clusters, "created": False}

    return create_response


def cluster_delete_clusters(clusters: list):

    cluster_ids = [c["cluster_id"] for c in clusters]
    logger.info(f"deleting clusters: {str(cluster_ids)}")

    for c in cluster_ids:
        try:
            cluster_action(c, ClusterAction.UNPIN)

        except Exception as e:
            logger.info(f"Warning: Failed to unpin cluster_id={c}")

        cluster_action(c, ClusterAction.DELETE)


def cluster_list() -> dict:

    response = _api_service.api_get(endpoint, "list")
    if response == {}:
        return None
    else:
        return response


def get_cluster_state(cluster_id: str):

    cluster = cluster_get(cluster_id)
    state = ClusterState[cluster["state"]]
    return state


def cluster_is_running(cluster_id: str):

    return get_cluster_state(cluster_id) == ClusterState.RUNNING


def cluster_is_terminated(cluster_id: str):

    return get_cluster_state(cluster_id) == ClusterState.TERMINATED


def cluster_get(cluster_id: str):

    query = f"cluster_id={cluster_id}"
    return _api_service.api_get(endpoint, "get", query=query)


def cluster_wait_until_state(
    cluster_id: str, cluster_state: ClusterState, wait_seconds: int = 10
):

    logger.info(
        f"Waiting for the cluster_id {cluster_id} to the reach the state: {cluster_state.name}"
    )

    if cluster_state not in [ClusterState.RUNNING, ClusterState.TERMINATED]:

        msg = f"Waiting for a cluster state {cluster_state.name} isn't final and isn't valid"
        logger.error(msg)
        raise Exception(msg)

    previous_state = ClusterState.UNKNOWN
    logger.info(f"The cluster_id {cluster_id} state: {previous_state.name}")

    while True:

        time.sleep(wait_seconds)

        cluster = cluster_get(cluster_id)

        state = ClusterState[cluster["state"]]
        # only log out if it's changed to reduce the log output
        if previous_state != state:
            logger.info(f"The cluster_id {cluster_id} state: {state.name}")
            previous_state = state

        if state == ClusterState.ERROR:

            msg = f"The cluster_id {cluster_id} is in error state: {cluster['state_message']}"
            logger.error(msg)
            raise Exception(msg)

        elif state == ClusterState.TERMINATED and cluster_state == ClusterState.RUNNING:

            msg = f"The cluster_id {cluster_id} {ClusterState.TERMINATED.name}: {cluster['state_message']}"
            logger.error(msg)
            raise Exception(msg)

        elif state == cluster_state:

            break


def cluster_has_tag(cluster: dict, tag_key: str, tag_value: str):

    if "custom_tags" in cluster:
        value = cluster["custom_tags"].get(tag_key)
        return value == tag_value
    else:
        return False


def clusters_clear_down(
    tag_key: str = None, tag_value: str = None, show_only: bool = True
):

    clusters = cluster_list()

    if tag_key and tag_value:
        clusters["clusters"] = [
            c for c in clusters["clusters"] if cluster_has_tag(c, tag_key, tag_value)
        ]
        logger.info(
            f"Deleting {len(clusters)} clusters found with tag {tag_key}={tag_value}"
        )
    else:
        logger.info(f"Deleting {len(clusters)} clusters found")

    if show_only:
        for c in clusters["clusters"]:
            cluster_id = c["cluster_id"]
            cluster = cluster_get(cluster_id)
            cluster_name = cluster["cluster_name"]
            spark_version = cluster["spark_version"]
            logger.info(
                f"name: {cluster_name} id: {cluster_id} spark_version: {spark_version} "
            )

    if not show_only:
        if clusters:
            for c in clusters["clusters"]:
                cluster_id = c["cluster_id"]
                cluster_action(cluster_id, ClusterAction.STOP)

            cluster_delete_clusters(clusters["clusters"])


def cluster_log_states():

    clusters = cluster_list()
    if clusters:
        for c in clusters["clusters"]:
            logger.info(f'{c["cluster_id"]}: {c["state"]}')
    else:
        logger.info("No clusters found")


def cluster_run(cluster_id: str):

    if cluster_is_terminated(cluster_id):
        logger.info(f"starting cluster {cluster_id}")
        cluster_action(cluster_id, ClusterAction.START)
        cluster_wait_until_state(cluster_id, ClusterState.RUNNING)

    elif not cluster_is_running(cluster_id):
        logger.info("cluster is starting")
        logger.info(f"cluster {cluster_id} is starting")
        cluster_wait_until_state(cluster_id, ClusterState.RUNNING)

    else:
        logger.info(f"cluster {cluster_id} is running")



import requests
from requests.exceptions import HTTPError
import logging
from typing import Union


logging.basicConfig(
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", level=logging.INFO
)
logger = logging.getLogger(f"autobricks.ApiUtils")


def base_api_get(
    url: str, headers: dict,  json: dict = None, data: dict = None, query: str = None
):

    if query:
        url = f"{url}?{query}"
    response = requests.get(url=url, headers=headers, json=json, data=data)

    try:
        response.raise_for_status()

    except HTTPError as e:

        msg = f"{e.response.status_code} error at {url} {e.response.text}"
        logger.error(msg)
        raise e

    return response


def base_api_post(url: str, headers: dict, json: dict = None, data: dict = None):

    response = requests.post(url=url, headers=headers, json=json, data=data)

    try:
        response.raise_for_status()

    except HTTPError as e:

        msg = f"{e.response.status_code} error at {url} {e.response.text}"
        logger.error(msg)
        raise e

    return response



from abc import ABC, abstractmethod
import logging
import adal
from .BaseApi import base_api_get as _api_get

logging.basicConfig(
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", level=logging.INFO
)
logger = logging.getLogger(f"autobricks.ApiUtils")

_AUTH_DNS = "login.microsoftonline.com"


class Auth(ABC):
    @abstractmethod
    def __init__(self, parameters: dict):
        pass

    @abstractmethod
    def get_headers(self):
        pass


class UserAuth(Auth):
    def __init__(self, parameters: dict):

        try:
            self.bearer_token = parameters["dbutilstoken"]
        except KeyError as e:
            raise KeyError("dbutilstoken key not found in UserAuth parameters")

    def get_headers(self):
        headers = {"Authorization": f"Bearer {self.bearer_token}"}
        return headers


class SPAuth(Auth):
    def __init__(self, parameters: dict):

        self.sp_client_id = parameters["sp_client_id"]
        self.sp_client_secret = parameters["sp_client_secret"]
        self.ad_resource = parameters["ad_resource"]
        self.tenant_id = parameters["tenant_id"]
        self._authority_url = f"https://{_AUTH_DNS}/{self.tenant_id}/oauth2/token"
        self._authority_headers = {"Content-Type": "application/x-www-form-urlencoded"}
        self._authority_data = {
            "grant_type": "client_credentials",
            "client_id": self.sp_client_id,
            "client_secret": self.sp_client_secret,
        }

        # get AD token
        self._authority_data["resource"] = self.ad_resource
        response = _api_get(
            url=self._authority_url,
            headers=self._authority_headers,
            data=self._authority_data,
        )
        self.bearer_token = response.json()["access_token"]

    def get_headers(self):
        headers = {"Authorization": f"Bearer {self.bearer_token}"}
        return headers


class SPMgmtEndpointAuth(SPAuth):
    def __init__(self, parameters: dict):

        # get AD token
        super().__init__(parameters)

        # get management endpoint token
        self.mgmt_resource_endpoint = parameters["mgmt_resource_endpoint"]
        self.workspace_name = parameters["workspace_name"]
        self.resource_group = parameters["resource_group"]
        self.subscription_id = parameters["subscription_id"]

        self._authority_data["resource"] = self.mgmt_resource_endpoint
        response = _api_get(
            url=self._authority_url,
            headers=self._authority_headers,
            data=self._authority_data,
        )
        self.mgmt_access_token = response.json()["access_token"]

    def get_headers(self):

        url = f"/subscriptions/{self.subscription_id}/resourceGroups/{self.resource_group}/providers/Microsoft.Databricks/workspaces/{self.workspace_name}"
        headers = {
            "Authorization": f"Bearer {self.bearer_token}",
            "X-Databricks-Azure-SP-Management-Token": self.mgmt_access_token,
            "X-Databricks-Azure-Workspace-Resource-Id": url,
        }
        return headers


class SPAdalAuth(Auth):
    def __init__(self, parameters: dict):

        self.sp_client_id = parameters["sp_client_id"]
        self.sp_client_secret = parameters["sp_client_secret"]
        self.ad_resource = parameters["ad_resource"]
        self.tenant_id = parameters["tenant_id"]

        self._authority_url = f"https://{_AUTH_DNS}/{self.tenant_id}"

        # get AD token
        context = adal.AuthenticationContext(self._authority_url)
        response = context.acquire_token_with_client_credentials(
            resource=self.ad_resource,
            client_id=self.sp_client_id,
            client_secret=self.sp_client_secret,
        )
        self.bearer_token = response["accessToken"]

    def get_headers(self):
        headers = {"Authorization": f"Bearer {self.bearer_token}"}
        return headers


class SPMgmtEndpointAdalAuth(SPAdalAuth):
    def __init__(self, parameters: dict):

        # get AD token
        super().__init__(parameters)
        self.mgmt_resource_endpoint = parameters["mgmt_resource_endpoint"]
        self.workspace_name = parameters["workspace_name"]
        self.resource_group = parameters["resource_group"]
        self.subscription_id = parameters["subscription_id"]

        # get management endpoint token
        context = adal.AuthenticationContext(self._authority_url)
        response = context.acquire_token_with_client_credentials(
            resource=self.mgmt_resource_endpoint,
            client_id=self.sp_client_id,
            client_secret=self.sp_client_secret,
        )
        self.mgmt_access_token = response["accessToken"]

    def get_headers(self):

        url = f"/subscriptions/{self.subscription_id}/resourceGroups/{self.resource_group}/providers/Microsoft.Databricks/workspaces/{self.workspace_name}"
        headers = {
            "Authorization": f"Bearer {self.bearer_token}",
            "X-Databricks-Azure-SP-Management-Token": self.mgmt_access_token,
            "X-Databricks-Azure-Workspace-Resource-Id": url,
        }
        return headers



from .ApiService import ApiService
from . import Configuration
from wheel_inspect import inspect_wheel
from dataclasses import dataclass
import os

_api_service = ApiService(Configuration.config)


def library_all_cluster_statuses():

    return _api_service.api_get("libraries", "all-cluster-statuses")


@dataclass
class Wheel:
    path: str
    name: str
    major: int
    minor: int
    patch: int
    dev: int


def _is_wheel(entry):

    ret = entry.path.endswith(".whl")
    ret = ret and entry.is_file()
    ret = ret and not ".dirty" in entry.path

    return ret


def _get_wheel_version(wheel_path: str):

    version = inspect_wheel(wheel_path)["version"].split(".")

    if len(version) > 3:
        dev = int(version[3].replace("dev", ""))
    else:
        dev = 0

    name = inspect_wheel(wheel_path)["project"]
    wheel = Wheel(wheel_path, name, version[0], version[1], version[2], dev)

    return wheel


def get_latest_wheel(path: str, name: str):

    directory = os.path.abspath(path)
    wheels = [
        _get_wheel_version(entry.path)
        for entry in os.scandir(directory)
        if _is_wheel(entry)
    ]
    wheels = [w for w in wheels if name == w.name]
    if len(wheels) > 0:
        wheels.sort(key=lambda x: (x.major, x.minor, x.patch, x.dev), reverse=True)
        wheel = wheels[0]
    else:
        wheel = None

    return wheel



from .DecodeUtils import (
    format_path_for,
    OS as OsEnum,
)
from ApiService import ApiService
from . import Configuration
from uuid import UUID
import time
import logging

endpoint = "jobs"
logging.basicConfig(
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)

_api_service = ApiService(Configuration.config)


def job_run_get(run_id: int):

    try:
        response = _api_service.api_get(endpoint, "runs/get", query=f"run_id={run_id}")
    except Exception as e:
        response = {"run_id": -1, "message": str(e)}

    return response


def job_runs_list():

    response = _api_service.api_get(endpoint, "runs/list")

    return response


def job_run_delete(run_id: int):
    data = {"run_id": run_id}
    try:
        response = _api_service.api_post(endpoint, "runs/delete", data)
    except Exception as e:
        response = {}

    return response


def job_run_submit(
    notebook_path: str,
    run_name: str = "default",
    spark_version: str = "7.3.x-scala2.12",
    node_type_id: str = "Standard_DS3_v2",
    driver_node_type_id: str = "Standard_DS3_v2",
    num_workers: int = 1,
    timeout_seconds: int = 900,
    idempotency_token: UUID = None,
    cluster_id: str = None,
):

    fmt_notebook_path = format_path_for(notebook_path, OsEnum.LINUX)

    data = {
        "notebook_task": {"notebook_path": fmt_notebook_path},
        "run_name": run_name,
        "libraries": [],
        "timeout_seconds": timeout_seconds,
    }

    if cluster_id:
        logger.info(
            f"Configuring job {fmt_notebook_path} to run on existing cluster {cluster_id}"
        )
        data["existing_cluster_id"] = cluster_id

    else:
        logger.info(
            f"Configuring job {fmt_notebook_path} to run on a new cluster {node_type_id}"
        )
        data["new_cluster"] = {
            "spark_version": spark_version,
            "node_type_id": node_type_id,
            "driver_node_type_id": driver_node_type_id,
            "num_workers": num_workers
            # "autotermination_minutes": autotermination_minutes
        }

    if idempotency_token:
        data["idempotency_token"] = str(idempotency_token)

    logger.info(f"Submitting job for notebook {fmt_notebook_path}")
    return _api_service.api_post(endpoint, "runs/submit", data)


def job_run_notebook(
    notebook_path: str,
    name: str = "default",
    idempotency_token: UUID = None,
    cluster_id: str = None,
    wait_seconds: int = 5,
):

    fmt_notebook_path = format_path_for(notebook_path, OsEnum.LINUX)

    start_time = time.time()
    response = job_run_submit(
        fmt_notebook_path,
        name,
        idempotency_token=idempotency_token,
        cluster_id=cluster_id,
    )

    run_id = response["run_id"]
    previous_state = {"run_id": run_id}

    logger.info(f"Notebook {fmt_notebook_path} job has start on run {run_id}")

    try:
        run_page_url = None

        while True:

            response = job_run_get(run_id)

            state = response["state"]
            life_cycle_state = response["state"]["life_cycle_state"]
            run_page_url = response["run_page_url"]

            if state != previous_state:
                logger.info(
                    f"Notebook:{fmt_notebook_path} State:{life_cycle_state} Url:{run_page_url}"
                )
                previous_state = state

            time.sleep(wait_seconds)

            if life_cycle_state == "TERMINATED":
                break

    except Exception as e:
        msg = f"Notebook:{fmt_notebook_path} State:ERROR Url:{run_page_url} Message:{str(e)}"
        logger.error(msg)
        raise Exception(msg)

    return state



from enum import Enum
from .Auth import (
    UserAuth,
    SPAuth,
    SPMgmtEndpointAuth,
    SPAdalAuth,
    SPMgmtEndpointAdalAuth,
)


class AuthenticationType(Enum):

    USER = (1,)
    SERVICE_PRINCIPLE = (2,)
    SERVICE_PRINCIPLE_MGMT_ENDPOINT = (3,)
    SERVICE_PRINCIPLE_ADAL = (4,)
    SERVICE_PRINCIPLE_MGMT_ENDPOINT_ADAL = 5


class AuthFactory:
    def __init__(self):
        self._creators = {}

    def register_format(self, auth_type: AuthenticationType, creator: type):
        self._creators[auth_type] = creator

    def get_auth(self, auth_type: AuthenticationType, parameters: dict):

        creator = self._creators.get(auth_type)
        if not creator:
            raise ValueError(auth_type)

        return creator(parameters)


auth_factory = AuthFactory()
auth_factory.register_format(AuthenticationType.USER, UserAuth)
auth_factory.register_format(AuthenticationType.SERVICE_PRINCIPLE, SPAuth)
auth_factory.register_format(
    AuthenticationType.SERVICE_PRINCIPLE_MGMT_ENDPOINT, SPMgmtEndpointAuth
)
auth_factory.register_format(AuthenticationType.SERVICE_PRINCIPLE_ADAL, SPAdalAuth)
auth_factory.register_format(
    AuthenticationType.SERVICE_PRINCIPLE_MGMT_ENDPOINT_ADAL, SPMgmtEndpointAdalAuth
)



import base64
import sys
from enum import Enum
import logging


logging.basicConfig(
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


class OS(Enum):

    WINDOWS = 1
    LINUX = 2
    MAC = 3


def base64_decode(base64_string: str, encoding: str = "utf-8"):

    base64_bytes = base64_string.encode(encoding)
    content_bytes = base64.b64decode(base64_bytes)

    return content_bytes


def base64_encode(string_bytes: bytes, ecoding: str = "utf-8"):

    base64_bytes = base64.b64encode(string_bytes)

    return base64_bytes.decode(ecoding)


def is_windows():

    return sys.platform in ["win32", "cygwin", "msys"]


def format_path_for(path: str, os: OS):

    if os in (OS.LINUX, OS.MAC):
        return path.replace("\\", "/")

    elif os == OS.WINDOWS:
        return path.replace("/", "\\")


def format_path_for_os(path: str):

    if sys.platform in ["win32", "cygwin", "msys"]:
        return format_path_for(path, OS.WINDOWS)

    elif sys.platform in ["linux", "linux2"]:
        return format_path_for(path, OS.LINUX)

    elif sys.platform in ["darwin"]:
        return format_path_for(path, OS.MAC)

    else:
        raise Exception(
            f"Error formating path={path} for os. Operating system not supported {sys.platform}"
        )



from .ApiService import ApiService

from .DecodeUtils import (
    base64_encode,
    base64_decode,
    format_path_for_os,
    format_path_for,
    OS,
)
from . import Configuration
import os
import logging

from enum import Enum

endpoint = "workspace"
logging.basicConfig(
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)

_api_service = ApiService(Configuration.config)


class DeployMode(Enum):
    DEFAULT = 1
    MOVE = 2
    PARENT = 3
    CHILD = 4
    ROOT_CHILD = 5


class Extension(Enum):
    py = 1
    html = 2
    ipynb = 3
    dbc = 4


class Format(Enum):
    SOURCE = 1
    HTML = 2
    JUPYTER = 3
    DBC = 4


class Language(Enum):
    SCALA = 1
    PYTHON = 2
    SQL = 3
    R = 4


def get_format(extension: Extension):

    if extension == Extension.py:
        return Format.SOURCE
    elif extension == Extension.html:
        return Format.HTML
    elif extension == Extension.ipynb:
        return Format.JUPYTER
    elif extension == Extension.dbc:
        return Format.DBC


def workspace_import(
    form_path: str, to_path: str, format: Format, language: Language, overwrite: True
):

    with open(form_path, "rb") as file:
        content = base64_encode(file.read())

    data = {
        "path": to_path,
        "format": format.name,
        "language": language.name,
        "content": content,
        "overwrite": overwrite,
    }
    return _api_service.api_post(endpoint, "import", data)


def workspace_delete(path: str, recursive: True):

    data = {"path": path, "recursive": recursive}
    return _api_service.api_post(endpoint, "delete", data)


def workspace_get_status(path: str):

    data = {"path": path}
    return _api_service.api_get(endpoint, "get-status", data)


def workspace_list(path: str):

    data = {"path": path}
    return _api_service.api_get(endpoint, "list", data)


def workspace_mkdirs(path: str):

    data = {"path": path}
    return _api_service.api_post(endpoint, "mkdirs", data)


def workspace_export(from_path: str, format: Format, to_path: str):

    data = {"path": from_path, "format": format.name.upper(), "direct_download": False}

    response = _api_service.api_get(endpoint, "export", data)

    file_type = response["file_type"]
    filename = os.path.basename(from_path)
    file_path = f"{to_path}/{filename}.{file_type}"

    content = base64_decode(response["content"])
    with open(file_path, "wb") as file:
        file.write(content)

    if file_type == "py":
        response = {
            "from_path": from_path,
            "to_path": file_path,
            "file_type": Format.SOURCE.name,
        }

    elif file_type == "html":
        response = {
            "from_path": from_path,
            "to_path": file_path,
            "file_type": Format.HTML.name,
        }

    elif file_type == "ipynb":
        response = {
            "from_path": from_path,
            "to_path": file_path,
            "file_type": Format.JUPYTER.name,
        }

    elif file_type == "dbc":
        response = {
            "from_path": from_path,
            "to_path": file_path,
            "file_type": Format.DBC.name,
        }

    return response


def workspace_dir_exists(path: str):

    try:
        reponse = workspace_get_status(path)
    except Exception as e:
        return False

    return reponse.get("object_type") == "DIRECTORY" and reponse.get("path") == path


def workspace_notebook_exists(path: str):

    try:
        reponse = workspace_get_status(path)
    except Exception as e:
        return False

    return reponse.get("object_type") == "NOTEBOOK" and reponse.get("path") == path


def workspace_import_dir(
    from_notebook_root: str,
    source_dir: str = "/",
    target_dir: str = None,
    deploy_mode: DeployMode = DeployMode.DEFAULT,
):

    if deploy_mode == DeployMode.DEFAULT and target_dir:

        logger.error(
            f"target_dir is not required for {deploy_mode.name} deployment mode"
        )
        raise Exception(
            f"target_dir is not required for {deploy_mode.name} deployment mode"
        )

    elif deploy_mode != DeployMode.DEFAULT and not target_dir:

        logger.error(f"target_dir is required for {deploy_mode.name} deployment mode")
        raise Exception(
            f"target_dir is required for {deploy_mode.name} deployment mode"
        )

    # automaticall adjust for user entered paths using the wrong separators
    source_dir = format_path_for_os(source_dir)
    from_root_path = os.path.abspath(format_path_for_os(from_notebook_root))

    response = {
        "from_notebook_root": from_notebook_root,
        "source_dir": source_dir,
        "target_dir": target_dir,
        "actions": [],
    }

    deploy_this = False

    logger.info(
        f"Start deploying from from_root_path={from_root_path} source_dir={source_dir} target_dir={target_dir} deploy_mode={deploy_mode.name}"
    )

    for root, subdirs, files in os.walk(from_root_path):

        deploy_dir = root.replace(from_root_path, "")

        logger.debug(
            f"Checking if deploy_dir={deploy_dir} starts with source_dir={source_dir}"
        )

        if deploy_dir.startswith(source_dir) or not source_dir or source_dir == "/":

            deploy_this = True
            if deploy_dir == "":
                deploy_dir = "/"

        else:
            deploy_this = False
            logger.info(
                f"Skipping dir source_dir={source_dir} deploy_dir={deploy_dir} target_dir={target_dir} deploy_mode={deploy_mode.name}"
            )

        if deploy_this:

            logger.info(
                f"Deploying dir source_dir={source_dir} deploy_dir={deploy_dir} target_dir={target_dir} deploy_mode={deploy_mode.name}"
            )

            action = _deploy_dir(source_dir, deploy_dir, target_dir, deploy_mode)
            if action:
                response["actions"].append(action)

            # deploy the notebooks
            for filename in files:

                from_file_path = os.path.join(root, filename)
                to_file_path = from_file_path.replace(from_root_path, "")

                logger.info(
                    f"Deploying from_file_path={from_file_path} source_dir={source_dir} deploy_dir={deploy_dir} target_dir={target_dir} deploy_mode={deploy_mode.name}"
                )

                action = _deploy_file(
                    from_file_path, source_dir, to_file_path, target_dir, deploy_mode
                )
                response["actions"].append(action)

    return response


def _deploy_file(
    from_file_path: str,
    source_dir: str,
    to_file_path: str,
    target_dir: str,
    deploy_mode: DeployMode,
):

    to_file_path = _modify_deploy_path(
        to_file_path, source_dir, target_dir, deploy_mode
    )

    to_file_path, extension = os.path.splitext(to_file_path)
    extension_type = Extension[extension.replace(".", "")]
    format = get_format(extension_type)
    language = Language.PYTHON
    overwrite = format != format.DBC

    action = {
        "action": "import",
        "from_file_path": from_file_path,
        "to_file_path": to_file_path,
        "format": format.name,
        "language": language.name,
        "overwrite": overwrite,
    }
    workspace_import(from_file_path, to_file_path, format, language, overwrite)

    return action


def _deploy_dir(
    source_dir: str, deploy_dir: str, target_dir: str, deploy_mode: DeployMode
):

    # insert the root sub-dir to support along side nested deployment if given.
    deploy_dir = _modify_deploy_path(deploy_dir, source_dir, target_dir, deploy_mode)

    # make the directory it's not the root
    action = None
    if deploy_dir != "/":
        action = {"action": "mkdirs", "path": deploy_dir}
        workspace_mkdirs(deploy_dir)

    return action


def _modify_deploy_path(
    deploy_dir: str, root: str, modifier: str, deploy_mode: DeployMode
):

    logger.info(f"Preparing deploy path deploy_dir={deploy_dir}")
    logger.debug(
        f"Preparing deploy path deploy_dir={deploy_dir} root={root} modifier={modifier} deploy_mode={deploy_mode.name}"
    )

    if deploy_mode == DeployMode.DEFAULT:
        new_path = deploy_dir

    elif deploy_mode == DeployMode.MOVE:
        new_path = deploy_dir.replace(root, modifier)

    elif deploy_mode == DeployMode.PARENT:
        modify_to = f"{modifier.replace('/.','')}{root}"
        new_path = deploy_dir.replace(root, modify_to)

    elif deploy_mode == DeployMode.CHILD:
        modify_to = f"{root}{modifier}"
        new_path = deploy_dir.replace(root, modify_to)

    elif deploy_mode == DeployMode.ROOT_CHILD:

        clean_modifier = modifier.replace("\\", "/")
        if clean_modifier.startswith("/"):
            clean_modifier = clean_modifier[1:]
        if clean_modifier.endswith("/"):
            clean_modifier = clean_modifier[:-1]

        folders = root.split("/")
        folders.insert(2, clean_modifier)
        modify_to = "/".join(folders)
        new_path = deploy_dir.replace(root, modify_to)

    # databricks runs on linux
    new_path = format_path_for(new_path, OS.LINUX)

    logger.info(f"Prepared deploy path deploy_dir={new_path}")

    return new_path



import distutils.cmd
import distutils.log
import os
from wheel_inspect import inspect_wheel
from dataclasses import dataclass
import shutil
import yaml


@dataclass
class Wheel:
    path: str
    name: str
    major: int
    minor: int
    patch: int
    dev: int


class DatabricksBuild(distutils.cmd.Command):

    description = "Build databricks project directory"
    user_options = [
        ("directory=", None, "path to databricks project directory"),
    ]

    def _copy_directory(self, from_path: str, to_path: str):

        try:
            # delete the destination dir if exists
            if os.path.exists(to_path):
                shutil.rmtree(to_path, ignore_errors=False, onerror=None)
            shutil.copytree(from_path, to_path)
        # Directories are the same
        except shutil.Error as e:
            print(f"Directory not copied. Error: {e}")
        # Any error saying that the directory doesn't exist
        except OSError as e:
            print(f"Directory not copied. Error: {e}")

    def _is_wheel(self, entry):

        ret = entry.path.endswith(".whl")
        ret = ret and entry.is_file()
        ret = ret and not ".dirty" in entry.path

        return ret

    def _get_wheel_version(self, wheel_path: str):

        version = inspect_wheel(wheel_path)["version"].split(".")

        if len(version) > 3:
            dev = int(version[3].replace("dev", ""))
        else:
            dev = 0

        name = inspect_wheel(wheel_path)["project"]
        wheel = Wheel(wheel_path, name, version[0], version[1], version[2], dev)

        return wheel

    def _get_latest_wheel(self, path: str, name: str = None):

        directory = os.path.abspath(path)
        wheels = [
            self._get_wheel_version(entry.path)
            for entry in os.scandir(directory)
            if self._is_wheel(entry)
        ]
        wheels = [w for w in wheels if name == w.name or not name]
        if len(wheels) > 0:
            wheels.sort(key=lambda x: (x.major, x.minor, x.patch, x.dev), reverse=True)
            wheel = wheels[0]
        else:
            wheel = None

        return wheel

    def initialize_options(self):

        """Set default values for options."""
        self.directory = "./databricks"
        self.dist = "./dist"
        self.databricks_dist_dir = f"{self.directory}/dist"

        if self.directory:
            assert os.path.exists(self.directory), (
                "Databricks project %s does not exist." % self.directory
            )

        if not os.path.exists(self.databricks_dist_dir):
            os.makedirs(self.databricks_dist_dir)

    def finalize_options(self):

        """Post-process options."""
        pass
        # if self.directory:
        #     assert os.path.exists(self.directory), (
        #         'Databricks project %s does not exist.' % self.directory)

    def _build_cluster_defns(
        self,
        cluster_defn_folder: str,
        semantic_version: str = "",
        init_script_path: str = None,
    ):

        directory = os.path.abspath(cluster_defn_folder)
        for entry in os.scandir(directory):
            if entry.path.endswith(".yaml") and entry.is_file():

                cluster_defn, init_scripts = self._build_cluster_defn(
                    entry.path, semantic_version, init_script_path
                )

                with open(entry.path, "w") as file:
                    data = yaml.dump(cluster_defn, file)

                if init_scripts:
                    self._build_init_scripts(init_scripts, entry.name)

    def _build_init_scripts(self, init_scripts: dict, cluster_defn_name: str):

        cluster_defn_name = cluster_defn_name.replace("yaml", "sh")
        from_path = f"{self.databricks_dist_dir}/init_scripts/{cluster_defn_name}"

        for i in init_scripts:

            if i.get("dbfs"):
                to_path = os.path.basename(i["dbfs"]["destination"])
                to_path = f"{self.databricks_dist_dir}/init_scripts/{to_path}"

                with open(from_path, "r") as from_file:
                    with open(to_path, "w") as to_file:
                        shell = from_file.readlines()

                        for l in shell:

                            if l[0:11] == "pip install":
                                path = l.split(" ")[-1]
                                filename = os.path.basename(path)
                                path = path.replace(filename, "")
                                filename, ext = os.path.splitext(filename)
                                wheel = self._get_latest_wheel(self.dist, filename)
                                if wheel:
                                    wheel_filename = os.path.basename(wheel.path)
                                    to_file.write(f"pip install {path}{wheel_filename}")
                                else:
                                    to_file.write(l)
                            else:
                                to_file.write(l)

                os.remove(from_path)

    def _build_cluster_defn(
        self,
        cluster_defn_path: str,
        semantic_version: str = "",
        init_script_path: str = None,
    ):

        with open(cluster_defn_path, "r") as f:
            cluster_defn: dict = yaml.safe_load(f)

        # cluster name variables replacements
        filename, _ = os.path.splitext(os.path.basename(cluster_defn_path))
        spark_version: str = cluster_defn["spark_version"]
        dbr = spark_version.split("-")[0].replace(".x", "")

        cluster_name: str = cluster_defn["cluster_name"]
        cluster_name = cluster_name.replace("{filename}", filename)
        cluster_name = cluster_name.replace("{dbr}", dbr)
        cluster_name = cluster_name.replace("{version}", semantic_version)
        cluster_defn["cluster_name"] = cluster_name

        # cluster log variable replacements
        if cluster_defn.get("cluster_log_conf"):

            log = cluster_defn["cluster_log_conf"].get("dbfs")

            if log:
                log["destination"] = log["destination"].replace(
                    "{cluster_name}", cluster_name
                )

        init_script = None
        # init script variable replacements
        if cluster_defn.get("init_scripts"):

            for i in cluster_defn["init_scripts"]:

                if i.get("dbfs"):

                    to_path = i["dbfs"]["destination"].replace(
                        "{cluster_name}", cluster_name
                    )
                    i["dbfs"]["destination"] = to_path

            init_script = cluster_defn["init_scripts"]

        return cluster_defn, init_script

    def run(self):

        """Run the databricks poject build."""
        self.announce("Building databrics project", level=distutils.log.INFO)

        # copy the directories in the source code home
        # to the dist folder. If the dist folder exists
        # then destroy and re-create it ensuring previous
        # builds are wiped out
        for entry in os.scandir(self.directory):
            if entry.is_dir() and entry.name != "dist":

                # set the destination dir
                to_path = f"{self.databricks_dist_dir}/{entry.name}"

                # copy and overwrite the directory
                self._copy_directory(entry.path, to_path)

        # build the cluster definitions off the copies created in dist
        self._build_cluster_defns(f"{self.databricks_dist_dir}/clusters")



import logging
from .AuthFactory import auth_factory, AuthenticationType
from .Auth import Auth
from .BaseApi import base_api_get as _base_api_get, base_api_post as _base_api_post


logging.basicConfig(
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", level=logging.INFO
)
logger = logging.getLogger(f"autobricks.ApiUtils")


API_VERSION = "2.0"


class ApiService:
    def __init__(self, configuration: dict):

        auth_type_str = configuration["auth_type"]
        self.host = configuration["databricks_api_host"]
        auth_type: AuthenticationType = AuthenticationType[auth_type_str]
        auth: Auth = auth_factory.get_auth(auth_type, configuration)

        self._headers = auth.get_headers()

    def api_get(self, api: str, function: str, data: dict = None, query: str = None):

        url = f"{self.host}/api/{API_VERSION}/{api}/{function}"
        if query:
            url = f"{url}?{query}"
        response = _base_api_get(url=url, headers=self._headers, json=data)

        return response.json()

    def api_post(self, api: str, function: str, data: dict):

        url = f"{self.host}/api/{API_VERSION}/{api}/{function}"
        response = _base_api_post(url=url, headers=self._headers, json=data)

        return response.json()



from .ApiService import ApiService
from .DecodeUtils import base64_decode, format_path_for_os
from . import Configuration
import base64
import os, fnmatch
import logging

endpoint = "dbfs"
logging.basicConfig(
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)

_api_service = ApiService(Configuration.config)


def dbfs_upload(from_path: str, to_path: str, overwrite: bool = True):

    # Create a handle that will be used to add blocks
    data = {"path": to_path, "overwrite": "true"}

    handle = _api_service.api_post(endpoint, "create", data)["handle"]

    with open(from_path, "rb") as f:

        for block in _read_file_block(f):

            # data = block.encode("utf-8")
            data = base64.b64encode(block)
            data = {"handle": handle, "data": data.decode()}
            _api_service.api_post(endpoint, "add-block", data)

    # close the handle to finish uploading
    data = {"handle": handle}
    return _api_service.api_post(endpoint, "close", data)


def dbfs_delete_file(path: str, recursive: bool = True):

    data = {"path": path, "recursive": recursive}
    return _api_service.api_post(endpoint, "delete", data)


def dbfs_get_status(path: str):

    data = {"path": path}
    return _api_service.api_get(endpoint, "get-status", data)


def dbfs_list(path: str):

    data = {"path": path}
    return _api_service.api_get(endpoint, "list", data)


def dbfs_mkdirs(path: str):

    data = {"path": path}
    return _api_service.api_post(endpoint, "mkdirs", data)


def dbfs_move(source_path: str, destination_path: str):

    data = {"source_path": source_path, "destination_path": destination_path}
    return _api_service.api_post(endpoint, "move", data)


def dbfs_read(path: str, offset: int, length: int):

    data = {"path": path, "offset": offset, "length": length}
    return _api_service.api_get(endpoint, "read", data)


def dbfs_download(from_path: str, to_path: str):

    to_os_path = format_path_for_os(to_path)

    logger.info(f"Starting to download file from dbfs: {from_path} => {to_os_path}")

    offset = 0
    # api limited to 1mb chunks
    read_chunk = 1024
    total_bytes = 0
    with open(to_os_path, "wb") as file:
        while True:

            response = dbfs_read(from_path, offset, read_chunk)
            offset += read_chunk
            bytes_read = response["bytes_read"]
            total_bytes += bytes_read
            data = response["data"]
            data = base64_decode(data)

            file.write(data)

            if bytes_read < read_chunk:
                break

    logger.info(f"Finished downloading {str(total_bytes)} bytes")


def _read_file_block(file_object, chunk_size=1024):
    while True:
        data = file_object.read(chunk_size)
        if not data:
            break
        yield data


def find_file(pattern: str, path: str):
    result = []
    for root, dirs, files in os.walk(path):
        for name in files:
            if fnmatch.fnmatch(name, pattern):
                result.append(os.path.join(root, name))
    return result



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "PYPI.md").read_text()

# This call to setup() does all the work
setup(
    name="autobricks",
    version_config={
        "template": "{tag}",
        "dev_template": "{tag}.dev{ccount}",
        "dirty_template": "{tag}.dev{ccount}.dirty",
        "starting_version": "0.0.1",
        "version_callback": None,
        "version_file": None,
        "count_commits_from_version_file": False
    },
    setup_requires=['setuptools-git-versioning'],
    description="Databricks Deployment Utils",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://autobricks.readthedocs.io/en/latest/",
    project_urls={
        'GitHub': 'https://github.com/semanticinsight/autobricks',
        'Documentation': 'https://autobricks.readthedocs.io/en/latest/'
    },
    author="Shaun Ryan",
    author_email="shaun_chiburi@hotmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    packages=["autobricks"],
    install_requires=[
          'requests',
          'PyYAML',
          'adal'
      ],
    zip_safe=False
)



import os

"""
Loads the configuration from environment variables.

"""

_default_schema = {
    "type": "object",
    "properties": {
        "auth_type": {
            "type": "string",
            "enum": [
                    "USER", 
                    "SERVICE_PRINCIPLE",
                    "SERVICE_PRINCIPLE_MGMT_ENDPOINT",
                    "SERVICE_PRINCIPLE_ADAL",
                    "SERVICE_PRINCIPLE_MGMT_ENDPOINT_ADAL"
                ]
            },
        "databricks_api_host": {"type": "string"},
    },
    "required": [
        "auth_type", 
        "databricks_api_host"
    ]
}

_user_schema = {
    "type": "object",
    "properties": {
        "auth_type": {"type": "string"},
        "dbutilstoken": {"type": "string"},
        "databricks_api_host": {"type": "string"},
    },
    "required": [
        "auth_type", 
        "dbutilstoken", 
        "databricks_api_host"
    ]
}

_sp_schema = {
    "type": "object",
    "properties": {
        "auth_type": {"type": "string"},
        "sp_client_id": {"type": "string"},
        "sp_client_secret": {"type": "string"},
        "ad_resource": {"type": "string"},
        "databricks_api_host": {"type": "string"},
    },
    "required": [
        "auth_type", 
        "sp_client_id", 
        "sp_client_secret",
        "ad_resource",
        "databricks_api_host"
    ]
}

_spme_schema = {
    "type": "object",
    "properties": {
        "auth_type": {"type": "string"},
        "sp_client_id": {"type": "string"},
        "sp_client_secret": {"type": "string"},
        "ad_resource": {"type": "string"},
        "tenant_id": {"type": "string"},
        "mgmt_resource_endpoint": {"type": "string"},
        "workspace_name": {"type": "string"},
        "resource_group": {"type": "string"},
        "subscription_id": {"type": "string"},
        "databricks_api_host": {"type": "string"},
    },
    "required": [
        "auth_type", 
        "sp_client_id", 
        "sp_client_secret",
        "ad_resource",
        "tenant_id",
        "mgmt_resource_endpoint",
        "workspace_name",
        "resource_group",
        "subscription_id",
        "databricks_api_host"
    ]
}

config = {
    "dbutilstoken": os.getenv("DBUTILSTOKEN"),
    "sp_client_id": os.getenv("SP_CLIENT_ID"),
    "sp_client_secret": os.getenv("SP_CLIENT_SECRET"),
    "ad_resource": os.getenv("AD_RESOURCE", "2ff814a6-3304-4ab8-85cb-cd0e6f879c1d"),
    "tenant_id": os.getenv("TENANT_ID"),
    "mgmt_resource_endpoint": os.getenv(
        "MGMT_RESOURCE_ENDPOINT", "https://management.core.windows.net/"
    ),
    "workspace_name": os.getenv("WORKSPACE_NAME"),
    "resource_group": os.getenv("RESOURCE_GROUP"),
    "subscription_id": os.getenv("SUBSCRIPTION_ID"),
    "auth_type": os.getenv("AUTH_TYPE", "SERVICE_PRINCIPLE"),
    "databricks_api_host": os.getenv("DATABRICKS_API_HOST"),
}
